public class Racional implements Cloneable {
	public Racional(final int numerador, final int denominador) {
		if (denominador == 0)
			throw new IllegalArgumentException("Denominador zero!");

		int mdc = mdc(numerador, denominador);

		if(denominador < 0) {
			this.numerador = -numerador / mdc;
			this.denominador = -denominador / mdc;
		} else {
			this.numerador = numerador / mdc;
			this.denominador = denominador / mdc;
		}
	}

	public Racional(final int valor) {
		numerador = valor;
		denominador = 1;
	}

	public Racional() {
		numerador = 0;
		denominador = 1;
	}

	protected void finalize() throws Throwable {
	}

	public String toString() {
		if (denominador == 1)
			return String.valueOf(numerador);
		else
			return numerador + "/" + denominador;
	}

	public int numerador() {
		return numerador;
	}

	public int denominador() {
		return denominador;
	}

	public void acrescentaDe(final Racional outro) {
		if (outro == null)
			throw new IllegalArgumentException("Refer�ncia para outro nula!");

		numerador =
			numerador * outro.denominador + outro.numerador * denominador;
		denominador *= outro.denominador;

		int mdc = mdc(numerador, denominador);

		numerador /= mdc;
		denominador /= mdc;
	}

	public static Racional somaDe(final Racional um, final Racional outro) {
		if (um == null)
			throw new IllegalArgumentException("Refer�ncia para um nula!");
		if (outro == null)
			throw new IllegalArgumentException("Refer�ncia para outro nula!");

		Racional soma = (Racional)um.clone();
		soma.acrescentaDe(outro);
		
		return soma;
	}

	private int numerador = 0;
	private int denominador = 1;

	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException exception) {
			throw new InternalError(exception.toString());
		}
	}

	public boolean equals(Object objecto) {
		if (this == objecto) {
		    return true;
		} else if(!objecto.getClass().equals(getClass())) {
		    return false;
		} else {
			Racional outro = (Racional) objecto;
			return numerador() == outro.numerador() && denominador() == outro.denominador();
		}
	}

	public boolean cumpreInvariante() {
		return 0 < denominador && mdc(numerador, denominador) == 1;
	}

	private static int mdc(final int m, final int n) {
		assert m != 0 || n != 0 : "m = " + m + " n = " + n;

		int a = Math.abs(m);
		int b = Math.abs(n);

		while (true) {
			if (a == 0) {
				assert 0 < b
					&& m % b == 0
					&& n % b == 0 : "m = " + m + ", n = " + n + " e mdc = " + b;
				return b;
			}

			b %= a;

			if (b == 0) {
				assert 0 < a
					&& m % a == 0
					&& n % a == 0 : "m = " + m + ", n = " + n + " e mdc = " + a;
				return a;
			}

			a %= b;
		}
	}
	
	public static final void main(String[] args) {
		assert Racional.mdc(0, 3) == 3 : "mdc(int, int)";
		assert Racional.mdc(3, 0) == 3 : "mdc(int, int)";
		assert Racional.mdc(2, 4) == 2 : "mdc(int, int)";
		assert Racional.mdc(10, 8) == 2 : "mdc(int, int)";
		assert Racional.mdc(-12, 8) == 4 : "mdc(int, int)";

		Racional r1 = new Racional(2, 4);
		Racional r2 = new Racional(2, 4);

		assert r1.numerador() == 1 : "numerador()";
		assert r1.denominador() == 2 : "denominador()";

		r1.denominador = 0;
		assert r1.equals(r1) : "equals(Object)";

		r2.denominador = 0;        
		r1.acrescentaDe(r2);

		assert r1.numerador() == 1 : "acrescentaDe(Racional)";
		assert r1.denominador() == 1 : "acrescentaDe(Racional)";

		Racional r3 = somaDe(r1, r2);

		assert r3.numerador() == 3 : "somaDe(Racional)";
		assert r3.denominador() == 2 : "somaDe(Racional)";

		assert new Racional(2).equals(new Racional(4, 2)) : "Racional(int)";
		assert new Racional().equals(new Racional(0)) : "Racional(int)";

		r3.denominador = 0;
		r3 = null;

		System.runFinalization();
		System.gc();
	}
}
